"use strict";

{
	const tempQuad = new C3.Quad();
	
	C3.Plugins.Rex_TweenTasks.Instance = class Rex_TweenTasksInstance extends C3.SDKInstanceBase
	{
		constructor(inst, properties)
		{
			super(inst);
			this.tasksMgr = new window.rexObjs.TweenTaskMgrKlass(this);
			this.exp_task = null;
			this.exp_fnPercentage = 0;
			this.exp_bindInstUID = -1;
			this.exp_bindInstTypeSID = null;

			this.my_timescale = -1.0;
			this._StartTicking();
		}
		
		Release()
		{
			super.Release();
			this.tasksMgr.CleanAll();
		}
		
		SaveToJson()
		{
			return {
				"tasksMgr": this.tasksMgr.saveToJSON(),
				"ts": this.my_timescale
			};
		}
		
		LoadFromJson(o)
		{
			this.tasksMgr.loadFromJSON(o["tasksMgr"]);
			this.my_timescale = o["ts"];
		}

		Tick()
		{
			var dt = this._runtime.GetDt(this._inst);
			if (dt === 0)
				return;

			this.tasksMgr.Tick(dt);
		}

		GetDebuggerProperties()
		{
			var prop = [];
			this.tasksMgr.getDebuggerValues(prop);

			return {
				"title": this.type.name,
				"properties": prop
			};
		};

		OnTaskStart(task) {
			this.exp_task = task;
			this.Trigger(C3.Plugins.Rex_TweenTasks.Cnds.OnAnyTaskStart);
			this.Trigger(C3.Plugins.Rex_TweenTasks.Cnds.OnTaskStart);
			this.exp_task = null;
		};
		OnTaskDone(task) {
			this.exp_task = task;
			this.Trigger(C3.Plugins.Rex_TweenTasks.Cnds.OnAnyTaskDone);
			this.Trigger(C3.Plugins.Rex_TweenTasks.Cnds.OnTaskDone);
			this.exp_task = null;
		};
		CallFunction(task, percentage, bindInstUID, bindInstTypeSID) {
			this.exp_task = task;
			this.exp_fnPercentage = percentage;
			this.exp_bindInstUID = bindInstUID;
			this.exp_bindInstTypeSID = bindInstTypeSID;
			this.Trigger(C3.Plugins.Rex_TweenTasks.Cnds.OnFunction);
			this.exp_task = null;
			this.exp_fnPercentage = 0;
			this.exp_bindInstUID = -1;
			this.exp_bindInstTypeSID = null;
		};
		PickBindInst(uid, sid) {
			var inst = this._runtime.GetInstanceByUID(uid);
			if (!inst)
				return;
			var objtype = this.get_objtype(sid);
			if (!objtype)
				return;

			var sol = objtype.getCurrentSol();
			sol.select_all = false;
			sol.instances.length = 0; // clear contents
			sol.instances.push(inst);
			objtype.applySolToContainer();
		};
		get_objtype(sid) {
			var idx, objtype;
			if (!_sid2typeIndex.hasOwnProperty(sid)) {
				idx = get_typeIndex(this._runtime.types_by_index, sid);
				_sid2typeIndex[sid] = idx;
				objtype = this._runtime.types_by_index[idx];
			} else {
				idx = _sid2typeIndex[sid];
				objtype = this._runtime.types_by_index[idx];
				if ((!objtype) || (objtype.sid != sid)) {
					delete _sid2typeIndex[sid];
					return this.get_objtype(sid);
				}
			}
			return objtype;
		};
		GetFnParam(paramName, valueType) {
			if (!this.exp_task)
				return 0;

			var fnParams = this.exp_task.fnParams;
			var v;
			if (fnParams.hasOwnProperty(paramName)) {
				var param = fnParams[paramName];
				if (param.updateFlag) // using default easeType
					easeParam(param, this.exp_fnPercentage, param.easeType);

				v = param.GetValue(valueType);
			} else
				v = 0;

			return v;
		};

	};
	
}
